@@ -8,7 +8,6 @@ from pulp_smash.pulp3.constants import (
     BASE_CONTENT_PATH,
 )
 
-AH_AUTH_URL = "https://sso.redhat.com/auth/realms/redhat-external/protocol/openid-connect/token"
 
 GALAXY_ANSIBLE_BASE_URL = "https://galaxy.ansible.com"
 
@@ -84,6 +83,6 @@ collections:
   version: ">=1.0.0,<=2.0.0"
   source: https://galaxy-dev.ansible.com/
 - testing.k8s_demo_collection
+- rshad.collection_demo
+- brightcomputing.bcm
 """
-
-AUTOMATION_HUB_URL = "https://cloud.redhat.com/api/automation-hub/"
